'use strict';
import knexfile from '../knexfile';
import knex from 'knex';
import fs from 'fs';
import path from 'path';
import coronerdUtils = require('../src/lib/coronerdUtils');
import queryUtils = require('../src/lib/queryUtils');
import summedEvents = require('../src/lib/summedEvents');
const {
  composeMetricSumQuery,
  composeMetricsTableBaseFilter,
  composeMetricDistributionQuery,
} = summedEvents;
const { parseQuery, composeBaseQueryBody, addTimestampRangeToBin } = queryUtils;

let db = null;

beforeAll(async () => {
  try {
    fs.unlinkSync(path.resolve('database', 'overview.test.db'));

    // get DB
    db = knex(knexfile);

    // migrate
    // suppress annoying knex messages when migrating
    const log = console.log;
    console.log = () => {
      return undefined;
    };
    await db.migrate.latest();
    console.log = log;
  } catch (e) {
    null;
  }
});

afterAll(async () => {
  try {
    fs.unlinkSync(path.resolve('database', 'overview.test.db'));
  } catch (e) {
    null;
  }
});

describe('queryUtils', () => {
  const coronerUrl = 'test_url';
  const token = 'test_token';
  const project = 'test_project';
  const metricGroup = 'test_metric_group';
  const filters = <any>[
    {
      timestamp: [
        ['at-least', 1622731335696],
        ['at-most', 1622760000000],
      ],
      metric1: ['equal', 'foo'],
      metric2: ['not-equal', 'bar'],
      metric3: ['contains', 'baz'],
    },
  ];
  const versionQuery =
    '{"group":["<%versionAttribute%>"],"fold":{"timestamp":[["bin", 32]]},"order":[{"name":";group","ordering":"descending"}],"offset":0,"limit":20}';
  const order: any = [{ name: 'test', ordering: 'ascending' }];

  test('composeMetricsTableBaseFilter()', async () => {
    // Mock the describe call on the metrics table
    jest.spyOn(coronerdUtils, 'fetchDescribe').mockImplementation(async function () {
      return [{ name: 'metric1' }, { name: 'metric2' }] as any;
    });

    const f = await composeMetricsTableBaseFilter(coronerUrl, token, project, filters);
    // it should use the provided timestamp filter
    expect(f).toMatchObject({ timestamp: filters[0].timestamp });
    // it should filter on metric1 and metric2
    expect(f).toMatchObject({
      metric1: filters[0].metric1,
      metric2: filters[0].metric2,
    });
    // it should NOT filter on metric3, because it is not returned by the describe call
    expect(f).not.toMatchObject({
      metric3: filters[0].metric3,
    });
  });

  test('composeMetricSumQuery()', async () => {
    // Mock the describe call on the metrics table
    // NOTE: metric3 is not a valid metric query, so it isn't returned by fetchDescribe
    jest.spyOn(coronerdUtils, 'fetchDescribe').mockImplementation(async function () {
      return [{ name: 'metric1' }, { name: 'metric2' }] as any;
    });

    const q = await composeMetricSumQuery(coronerUrl, token, project, filters, metricGroup);

    // it should create a valid metric group filter query
    expect(q).toMatchObject({
      filter: [
        {
          _metric_group: [['equal', metricGroup]],
          timestamp: filters[0].timestamp,
          metric1: filters[0].metric1,
          metric2: filters[0].metric2,
        },
      ],
      group: ['*'],
      fold: { _value: [['sum']] },
      table: 'metrics',
    });
    // metric3 was not returned by our fetchDescribe mock
    expect(q?.filter?.[0]).not.toMatchObject({
      metric3: filters[0].metric3,
    });
  });

  test('composeMetricDistributionQuery()', async () => {
    // Mock the describe call on the metrics table
    // NOTE: metric3 is not a valid metric query, so it isn't returned by fetchDescribe
    jest.spyOn(coronerdUtils, 'fetchDescribe').mockImplementation(async function () {
      return [{ name: 'metric1' }, { name: 'metric2' }] as any;
    });

    const q = await composeMetricDistributionQuery(coronerUrl, token, project, filters);

    // it should create a valid metric group filter query
    expect(q).toMatchObject({
      filter: composeMetricsTableBaseFilter(coronerUrl, token, project, filters),
      group: ['_metric_group'],
      fold: { _value: [['sum']] },
      table: 'metrics',
    });
  });

  test('composeBaseQueryBody()', async () => {
    const q = composeBaseQueryBody(versionQuery, filters, order, 'test_attr', undefined, []);
    // finds and replaces version attribute strings
    expect(q.group).toMatchObject(['test_attr']);
    // adds filters and order
    expect(q.order).toBe(order);
    expect(q.filter).toBe(filters);

    // override query grouping if supplied
    const q2 = composeBaseQueryBody(versionQuery, filters, order, 'test_attr', ['my_group'], []);
    expect(q2.group).toMatchObject(['my_group']);
    // adds optional range params to timestamp bin agg if present
    expect(q2?.fold?.timestamp).toMatchObject([['bin', 32, 1622731335696, 1622760000000]]);

    // doesn't touch timestamp bin agg when time range not present in filter
    const q3 = composeBaseQueryBody(versionQuery, [] as any, order, 'test_attr', ['my_group'], []);
    expect(q3?.fold?.timestamp).toMatchObject([['bin', 32]]);
  });
  test('parseQuery()', async () => {
    parseQuery(versionQuery, undefined);
    // finds and replaces version attribute strings
    expect(parseQuery(versionQuery, 'test_attr').group).toMatchObject(['test_attr']);
    // does not find and replace when version attr is undefined
    expect(parseQuery(versionQuery, undefined).group).toMatchObject(['<%versionAttribute%>']);
  });

  test('addTimestampRangeToBin()', async () => {
    const q = parseQuery(versionQuery, undefined);
    // doesn't touch timestamp bin agg when time range not present in filter
    addTimestampRangeToBin(q, 'timestamp', [] as any);
    expect(q.fold?.timestamp).toMatchObject([['bin', 32]]);

    // adds optional range params to timestamp bin agg if present in filters
    addTimestampRangeToBin(q, 'timestamp', filters);
    expect(q.fold?.timestamp).toMatchObject([['bin', 32, 1622731335696, 1622760000000]]);
  });
});
